#include <iostream>
#include <cstdlib>
#include <signal.h>
#include "RtMidi.h"
#include <unistd.h>
#include <cstring>
#include <stdint.h>
#include <cstdio>

using namespace std;

inline void Sleep(uint32_t millisecs)
{
   if(millisecs % 1000)
   {
      usleep((millisecs % 1000)*1000);
   }
   sleep(millisecs/1000);
}

double elapsed()
{
   static timespec start;
   static bool bOnce = true;
   if(bOnce)
   {
      bOnce = false;
      clock_gettime(CLOCK_REALTIME,&start);
   }
   timespec ts;
   clock_gettime(CLOCK_REALTIME,&ts);
   double secs = (ts.tv_sec - start.tv_sec)*1000;
   double nsecs = ts.tv_nsec - start.tv_nsec;
   secs += nsecs/(1000*1000);
   return secs;
}
double setStart = elapsed();

int linux_kbhit()
{
   struct timeval tv;
   fd_set read_fd;

   tv.tv_sec=0;
   tv.tv_usec=0;
   FD_ZERO(&read_fd);
   FD_SET(0,&read_fd);

   if(select(1, &read_fd, NULL, NULL, &tv) == -1)
   return 0;

   if(FD_ISSET(0,&read_fd))
   return 1;

   return 0;
}

int WriteVarLen(uint32_t value, uint8_t (&outstr)[5])
{
   if(value > 0xFFFFFFF)
   {
      printf("TOO BIG!\n");
      outstr[0] = 0;
      return 0;
   }
   uint32_t buffer;
   buffer = value & 0x7F;

   while ( (value >>= 7) )
   {
     buffer <<= 8;
     buffer |= ((value & 0x7F) | 0x80);
   }

   int written = 0;
   while(true)
   {
      outstr[written++] = (uint8_t) buffer;
      if (buffer & 0x80)
          buffer >>= 8;
      else
          break;
   }
   outstr[written] = 0;
   return written;
}

/*uint32_t ReadVarLen(uint8_t *instring)
{
    uint32_t value = 0;
    uint8_t c;
    uint8_t *cur_char = instring;

    if((value = *cur_char++) & 0x80)
    {
       value &= 0x7F;
       do
       {
         value = (value << 7) + ((c = *cur_char++) & 0x7F);
       }
       while (c & 0x80);
    }

    return value;
}*/

//look at these links for explanation:
//http://www.skytopia.com/project/articles/midi.html
//http://midi.mathewvp.com/aboutMidi.htm
//http://web.archive.org/web/20070622094111/jedi.ks.uiuc.edu/~johns/links/music/midifile.html
//http://www.midimountain.com/midi/midi_status.htm

static int counter = 0;
static FILE *mf = 0;
static FILE *thelog = 0;
uint32_t track_bytes = 0;
static double last_rcvd_ms = -10000;
static double last_rcvd_timing_ms = -10000;
#define ACTIVITY_TIMEOUT_MS (1000*60*10) //Close out the file after 10 minutes
#define NO_TIMING_TIMEOUT_MS 2500 //If the piano is turned off, close the file after 2.5 seconds. Some startup delays can take 2 full seconds.
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static double deltatime_sum = 0;
static bool bAskedToStop = false;
static bool gotNoteOn = false;
static char fname[200];
static char log_dir[500];
static char port_name[200];

//This should give us about 37 hours max delta value, and 1/2 ms resolution with the default 120 beats per minute. Pretty standard.
static const uint16_t TICKS_PER_BEAT = 960; //two beats per second

//#define DO_LOGGING

static void finish(int ignore)
{
   bAskedToStop = true;
}

void open_mf()
{
   char fmtStr[] = "%Y%m%d__%H%M_%S";
   char timestr[200];
   time_t t = time(0);
   tm *today = localtime(&t);
   strftime(timestr,sizeof(timestr),fmtStr,today);
   sprintf(fname, "%s/MidiLog_%s.mid",log_dir,timestr);
   mf = fopen(fname,"wb");
   if(!mf)
   {
      cout << "Error opening " << fname << " for writing!" << endl;
      exit(1);
   }
   uint8_t ticks_high = (uint8_t)(TICKS_PER_BEAT >> 8);
   uint8_t ticks_low = (uint8_t)TICKS_PER_BEAT;

   //meaning:          M     T     h     d     header size-----------  type------  numtracks-
   uint8_t header[] = {0x4D, 0x54, 0x68, 0x64, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x01,
                       //ticks/beat---------  M     T     r     k---  Bytes following-------  All notes off---------
                       ticks_high, ticks_low, 0x4D, 0x54, 0x72, 0x6B, 0x00, 0x00, 0x00, 0x00, 0x00, 0xB0, 0x7B, 0x00};

   fwrite(header, 1, sizeof(header), mf);
   track_bytes = 4;
   gotNoteOn = false;

#ifdef DO_LOGGING
   char logname[200];
   sprintf(logname,"%s/full_log_%s",log_dir,timestr);
   thelog = fopen(logname,"wb");
   if(!thelog)
   {
      cout << "Error opening " << logname << " for writing!" << endl;
      exit(1);
   }
#endif
}

void callback_func(double deltatime, std::vector<uint8_t> *message_arg, void *userData)
{
   pthread_mutex_lock(&mutex);
   double NowTime_ms = elapsed();
   vector <unsigned char> &message = (*message_arg);
   deltatime_sum += deltatime;
   last_rcvd_timing_ms = NowTime_ms;
   switch(message[0])
   {
   case 0xFE:
   case 0xF8:
   case 0xFA:
   case 0xFC:
      pthread_mutex_unlock(&mutex);
      return;
   }

   static vector<uint8_t> orig_msg;
#ifdef DO_LOGGING
   orig_msg = message;
#endif
   static double first_rcvd_90_ms;
   static uint64_t total_deltas;

   if(!mf)
   {
      open_mf();
      total_deltas = 0;
      first_rcvd_90_ms = 0; //Never insert delays into the midi file until a Note On has come in.
   }

   static uint8_t last_status_byte = 0;

   //for sysex messages, add a size field
   if(message[0] == 0xF0)
   {
      vector<uint8_t>::iterator it = message.begin();
      it++;
      uint8_t size_field = (uint8_t)(message.size() - 1); //this should technically be a variable length quantity but in practice the P80 always sends small sysex's.
      message.insert(it,size_field);
      last_status_byte = 0;
   }
   else
   {
      if(message[0] == 0x90)
      {
         gotNoteOn = true;
         if(!first_rcvd_90_ms)
         {
            first_rcvd_90_ms = NowTime_ms - 100; //Allow a slight delay after sending a bunch of control messages?
         }
      }
      if(message[0] == last_status_byte)
      {
         message.erase(message.begin());
      }
      else
      {
         last_status_byte = message[0];
      }
   }

   uint8_t delta[5];
   uint32_t current_delta = 0;
   if(first_rcvd_90_ms)
   {
      current_delta = (uint32_t)((NowTime_ms - first_rcvd_90_ms)/1000.0*TICKS_PER_BEAT*2.0 - total_deltas);
   }
   total_deltas += current_delta;
   int delta_len = WriteVarLen(current_delta, delta); //deltatime_sum
   fwrite(delta,1,delta_len,mf);
   fwrite(&(message[0]),1,message.size(),mf);
   track_bytes += message.size() + delta_len;

#ifdef DO_LOGGING
   unsigned int nBytes = orig_msg.size();
   int printed = 0;
   char outbuf[300];
   printed += sprintf(outbuf,"%07i ",counter++);
   for(unsigned int i=0; i < nBytes; i++)
   {
      printed += sprintf(outbuf + printed,"%X ",(int)orig_msg[i]);
   }
   while(printed < 37)
   {
      outbuf[printed++] = ' ';
   }
   if(nBytes > 0)
   {
      printed += sprintf(outbuf + printed,"rcvd %.0f ms, delta = %.4f\n", NowTime_ms, deltatime_sum);
   }
   fwrite(outbuf, 1, printed, thelog);
   deltatime_sum = 0;
#endif

   last_rcvd_ms = NowTime_ms;
   pthread_mutex_unlock(&mutex);
}

void close_mf()
{
   uint8_t tag_close[] = {0x00, 0xB0, 0x7B, 0x00, 0x00, 0xFF, 0x2F, 0x00};
   fwrite(tag_close, 1, 8, mf);
   track_bytes += 8;
   fseek(mf, 18, SEEK_SET);
   for(int i = 3; i >= 0; i--)
   {
      char c = (char) (track_bytes >> (i*8));
      fwrite(&c, 1, 1, mf);
   }
   fclose(mf);
   if(!gotNoteOn)
   {
      cout << "Deleting a spurious file: " << fname << endl;
      unlink(fname);
   }
   mf = 0;
#ifdef DO_LOGGING
   fclose(thelog);
#endif
}

RtMidiIn *attempt_connection(bool initial)
{
   RtMidiIn *midiin = 0;

   try
   {
      midiin = new RtMidiIn();
   }
   catch(RtError &error)
   {
      error.printMessage();
      exit( EXIT_FAILURE );
   }

   // Check available ports vs. specified.
   int port = -1;
   uint32_t nPorts = midiin->getPortCount();

   string portName;
   for(uint32_t i = 0; i < nPorts; i++)
   {
      try
      {
         portName = midiin->getPortName(i);
      }
      catch(RtError &error)
      {
         error.printMessage();
         delete midiin;
         exit( EXIT_FAILURE);
      }
      //printf("Port name: %s\n",portName.c_str());
      //if(!(strcmp(portName.c_str(),"MidAir:0")))
      //if(!(strcmp(portName.c_str(),"MidiSport 2x2:0")))
      if(!(strcmp(portName.c_str(),port_name)))
      {
         port = (int) i;
      }
   }
   if(port == -1)
   {
      if(initial)
      {
         cout << "Could not find " << port_name << " port!" << endl;
      }
      delete midiin;
      exit(1);
   }

   try
   {
      midiin->openPort((uint32_t)port);
   }
   catch (RtError &error)
   {
      error.printMessage();
      delete midiin;
      exit( EXIT_FAILURE);
   }

   signal(SIGINT, finish);

   // Set our callback function.  This should be done immediately after
   // opening the port to avoid having incoming messages written to the
   // queue instead of sent to the callback function.
   midiin->setCallback(callback_func);

   // Don't ignore sysex, timing, or active sensing messages.
   midiin->ignoreTypes( false, false, false );

   return midiin;
}

int main( int argc, char *argv[] )
{
   if(argc < 3)
   {
      cout << "Usage: " << argv[0] << " LOG_DIR PORT_NAME" << endl;
      exit(1);
   }
   sprintf(log_dir,"%s",argv[1]);
   sprintf(port_name,"%s",argv[2]);

   RtMidiIn *midiin = 0;

   midiin = attempt_connection(true);
   if(!midiin)
   {
      exit(EXIT_FAILURE);
   }
   cout << "Reading MIDI input... Ctrl-C to quit." << endl;

   /*
   open_mf();

   uint8_t delta[5];
   int delta_len = WriteVarLen(TICKS_PER_BEAT*2 - 150, delta);

   uint8_t delta2[5];
   int delta2_len = WriteVarLen(150, delta2);

   uint8_t note_on[] = {0x3C, 0x60};
   uint8_t note_off[] = {0x3C, 0x0};
   uint8_t first_seq[] = {0x0, 0x90};
   fwrite(first_seq, 1, 2, mf);
   fwrite(note_on, 1, 2, mf);
   track_bytes += 4;
   for(int i = 0; i < 500; i++)
   {
      fwrite(delta2,1,delta2_len,mf);
      fwrite(note_off,1,2,mf);
      fwrite(delta,1,delta_len,mf);
      fwrite(note_on,1,2,mf);
      track_bytes += 4 + delta_len + delta2_len;
   }
   */
   while(!bAskedToStop)
   {
      pthread_mutex_lock(&mutex);
      double NowTime_ms = elapsed();
      if(mf && ((NowTime_ms > last_rcvd_ms + ACTIVITY_TIMEOUT_MS) || (NowTime_ms > last_rcvd_timing_ms + NO_TIMING_TIMEOUT_MS)))
      {
         close_mf();
      }
      pthread_mutex_unlock(&mutex);

      if(midiin)
      {
         if(midiin->FatalError)
         {
            cout << "Port was closed, attempting to reopen..." << endl;
            delete midiin;
            midiin = 0;
         }
      }
      else
      {
         midiin = attempt_connection(false);
         if(!midiin)
         {
            Sleep(1000);
         }
         else
         {
            cout << "Found connection again!" << endl;
         }
      }
      Sleep(100);
   }

   cout << "\nCleaning up..." << endl;
   if(midiin)
   {
      delete midiin;
   }
   if(mf)
   {
      close_mf();
   }

   return 0;
}
